/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ------------------------- */
/* BIG arithmetic operations */
/* ------------------------- */

extern  int     BigPrecision;

class   BIG     {

	protected:

	long x [MAX_BIG_PRECISION];

	public:

	inline  BIG () { };                     // Dummy constructor

	inline  BIG (BIG &a) {

		bcopy (a.x, x, sizeof (long) * BigPrecision);
	};

	friend  void    initBig ();             // Initializer

/* ----------- */
/* Conversions */
/* ----------- */

#if     ZZ_ATT
		BIG     (int);
		BIG     (long);
#else
	inline  BIG      (long i)        {       // Conversion from long

		x [0] = i;
		for (int k = 1; k < BigPrecision; x [k++] = 0L);

	};
#endif
	BIG     (double);               // Conversion from double/float

	operator double () const;       // Conversion to float/double

	// To check convertibility to double
	friend  int     convertible (const BIG&);
	// Internal
	friend  void    BShift (BIG&, int);

/* ----------------------------------------- */
/* Basic arithmetic relations and operations */
/* ----------------------------------------- */

#if     ZZ_ATT
	friend int operator== (const BIG&, const BIG&);
	friend int operator!= (const BIG&, const BIG&);
#else
	friend inline int  operator== (const BIG &a, const BIG &b) {
		for (int k = 0; k < BigPrecision; k++)
			if (a.x [k] != b.x [k]) return (0);
		return (1);
	};

	friend inline int  operator!= (const BIG &a, const BIG &b) {
		for (int k = 0; k < BigPrecision; k++)
			if (a.x [k] != b.x [k]) return (1);
		return (0);
	};
#endif
	friend  inline  int  operator== (const BIG &a, long b) {
		return (a == (BIG) b);
	};

	friend  inline  int  operator!= (const BIG &a, long b) {
		return (a != (BIG) b);
	};

	friend  inline  int  operator== (long a, const BIG &b) {
		return ((BIG) a == b);
	};

	friend  inline  int  operator!= (long a, const BIG &b) {
		return ((BIG) a != b);
	};
#if     ZZ_ATT
	friend  inline  int  operator== (const BIG &a, int b) {
		return (a == (BIG) b);
	};

	friend  inline  int  operator!= (const BIG &a, int b) {
		return (a != (BIG) b);
	};

	friend  inline  int  operator== (int a, const BIG &b) {
		return ((BIG) a == b);
	};

	friend  inline  int  operator!= (int a, const BIG &b) {
		return ((BIG) a != b);
	};
#endif

	friend  inline  int  operator== (const BIG &a, double b) {
		return (a == (BIG) b);
	};

	friend  inline  int  operator!= (const BIG &a, double b) {
		return (a != (BIG) b);
	};

	friend  inline  int  operator== (double a, const BIG &b) {
		return ((BIG) a == b);
	};

	friend  inline  int  operator!= (double a, const BIG &b) {
		return ((BIG) a != b);
	};

	friend  int     operator<  (const BIG&, const BIG&);   // Inequalities
	friend  int     operator<= (const BIG&, const BIG&);
	friend  int     operator>  (const BIG&, const BIG&);
	friend  int     operator>= (const BIG&, const BIG&);

	friend  inline int operator< (const BIG &a, long b) {
		return (a < (BIG) b);
	};

	friend  inline int operator< (long a, const BIG &b) {
		return ((BIG) a < b);
	};
#if     ZZ_ATT
	friend  inline int operator< (const BIG &a, int b) {
		return (a < (BIG) b);
	};

	friend  inline int operator< (int a, const BIG &b) {
		return ((BIG) a < b);
	};
#endif

	friend  inline int operator< (const BIG &a, double b) {
		return (a < (BIG) b);
	};

	friend  inline int operator< (double a, const BIG &b) {
		return ((BIG) a < b);
	};

	friend  inline int operator<= (const BIG &a, long b) {
		return (a <= (BIG) b);
	};

	friend  inline int operator<= (long a, const BIG &b) {
		return ((BIG) a <= b);
	};
#if     ZZ_ATT
	friend  inline int operator<= (const BIG &a, int b) {
		return (a <= (BIG) b);
	};

	friend  inline int operator<= (int a, const BIG &b) {
		return ((BIG) a <= b);
	};
#endif

	friend  inline int operator<= (const BIG &a, double b) {
		return (a <= (BIG) b);
	};

	friend  inline int operator<= (double a, const BIG &b) {
		return ((BIG) a <= b);
	};

	friend  inline int operator> (const BIG &a, long b) {
		return (a > (BIG) b);
	};

	friend  inline int operator> (long a, const BIG &b) {
		return ((BIG) a > b);
	};
#if     ZZ_ATT
	friend  inline int operator> (const BIG &a, int b) {
		return (a > (BIG) b);
	};

	friend  inline int operator> (int a, const BIG &b) {
		return ((BIG) a > b);
	};
#endif

	friend  inline int operator> (const BIG &a, double b) {
		return (a > (BIG) b);
	};

	friend  inline int operator> (double a, const BIG &b) {
		return ((BIG) a > b);
	};

	friend  inline int operator>= (const BIG &a, long b) {
		return (a >= (BIG) b);
	};

	friend  inline int operator>= (long a, const BIG &b) {
		return ((BIG) a >= b);
	};
#if     ZZ_ATT
	friend  inline int operator>= (const BIG &a, int b) {
		return (a >= (BIG) b);
	};

	friend  inline int operator>= (int a, const BIG &b) {
		return ((BIG) a >= b);
	};
#endif

	friend  inline int operator>= (const BIG &a, double b) {
		return (a >= (BIG) b);
	};

	friend  inline int operator>= (double a, const BIG &b) {
		return ((BIG) a >= b);
	};

	friend  BIG     operator+ (const BIG&, const BIG&); // Addition
	friend  BIG     operator+ (const BIG&, long);       // Addition of int
	friend  double  operator+ (const BIG&, double);     // Addition of float

	friend  inline BIG operator+ (long a, const BIG &b) {
		return (b + a);
	};
#if     ZZ_ATT
	friend  inline  BIG     operator+ (const BIG &a, int b) {
		return (a + (long)b);
	};
	friend  inline BIG operator+ (int a, const BIG &b) {
		return (b + (long)a);
	};
#endif

	friend  inline double operator+ (double a, const BIG &b) {
		return (b + a);
	};

	friend  BIG     operator- (const BIG&, const BIG&);  // Subtraction
	friend  BIG     operator- (const BIG&, long);   // Subtraction of int
	friend  double  operator- (const BIG&, double); // Of float
	friend  double  operator- (double, const BIG&); // From float
#if     ZZ_ATT
	friend  inline  BIG     operator- (const BIG &a, int b) {
		return (a - (long)b);
	};
#endif

	inline  BIG     operator++ () {                 // Autoincrement
		return (*this = *this + 1);
	};

	inline  BIG     operator-- () {                 // Autodecrement
		return (*this = *this - 1);
	};

	inline  BIG     operator+= (BIG a) {
		return (*this = *this + a);
	};

	inline  BIG     operator+= (long a) {
		return (*this = *this + a);
	};

	inline  BIG     operator-= (BIG a) {
		return (*this = *this - a);
	};

	inline  BIG     operator-= (long a) {
		return (*this = *this - a);
	};
#if     ZZ_ATT
	inline  BIG     operator+= (int a) {
		return (*this = *this + a);
	};
	inline  BIG     operator-= (int a) {
		return (*this = *this - a);
	};
#endif

	friend  BIG     operator* (const BIG&, const BIG&);  // Multiplication
	friend  BIG     operator* (const BIG&, long);        // By int
	friend  double  operator* (const BIG&, double);      // By float

	friend  inline BIG operator* (long a, const BIG &b) {
		return (b * a);
	};
#if     ZZ_ATT
	friend  inline  BIG     operator* (const BIG &a, int b) {
		return (a * (long) b);
	};
	friend  inline BIG operator* (int a, const BIG &b) {
		return (b * (long) a);
	};
#endif

	friend  inline double operator* (double a, const BIG &b) {
		return (b * a);
	};

	friend  BIG     operator/ (const BIG&, const BIG&);   // Division
	friend  double  operator/ (const BIG&, double);       // By float
	friend  double  operator/ (double, const BIG&);       // Of float

	friend  inline BIG operator/ (const BIG &a, long b) {
		return (a / (BIG) b);
	};

	friend  BIG     operator% (const BIG&, const BIG&);   // Modulo
	friend  long    operator% (const BIG&, long);         // By long

#if     ZZ_ATT
	friend  inline BIG operator/ (const BIG &a, int b) {
		return (a / (BIG) b);
	};
	friend  inline  int    operator% (const BIG &a, int b) {
		return ((int)(a % (long) b));
	};
#endif

	inline  BIG     operator*= (BIG a) {

		return (*this = *this * a);
	};

	inline  BIG     operator*= (long a) {

		return (*this = *this * a);
	};

	inline  BIG     operator/= (BIG a) {

		return (*this = *this / a);
	};

	inline  BIG     operator/= (long a) {

		return (*this = *this / a);
	};
#if     ZZ_ATT
	inline  BIG     operator*= (int a) {

		return (*this = *this * (long) a);
	};
	inline  BIG     operator/= (int a) {

		return (*this = *this / (long) a);
	};
#endif

	inline  BIG     operator%= (BIG a) {

		return (*this = *this % a);
	};

	inline  BIG     operator%= (long a) {

		return (*this = *this % a);
	};
#if     ZZ_ATT
	inline  BIG     operator%= (int a) {

		return (*this = *this % (long) a);
	};
#endif
	
	friend  BIG     ctob (char*);           // String to BIG conversion

	friend  char    *btoa (const BIG&, char*, int nc = 15); // Vice versa

};

extern  BIG   BIG_0, BIG_1, BIG_INF;            // Some constants of type big
