/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ----------------------------------- */
/* Operations on the object dictionary */
/* ----------------------------------- */

#include        "types.h"
#include        "var.h"

TItem::TItem (char *on, char *sn, char *bn, char *nn) {

/* -------------------------------------- */
/* The constructor for a dictionary entry */
/* -------------------------------------- */

	TItem   *ti;
	int     ix;

	// Find the owner
	if ((ti = findObj (on)) == NULL)
		// This shouldn't really happen
		abtDisplay (form ("%s: object not found (internal error)", on));

	// Append to the child list
	if ((lnext = ti->ChList) != NULL) lnext->lprev = this;
	lprev = (TItem*) (&(ti->ChList));
	ti->ChList = this;

	// Append to the hash list
	if ((hnext = HTable [ix = hash (sn)]) != NULL) hnext->hprev = this;
	hprev = (TItem*) ((&(HTable[ix])) - 2);
	HTable [ix] = this;

	// Allocate memory for the standard name
	SName = new char [strlen (sn) + 1];
	strcpy (SName, sn);

	if (bn [0] != '\0') {
		// Same for the base name
		BName = new char [strlen (bn) + 1];
		strcpy (BName, bn);
	} else
		BName = NULL;

	if (nn [0] != '\0') {
		// And for the nickname
		NName = new char [strlen (nn) + 1];
		strcpy (NName, nn);
	} else
		NName = NULL;

	ChList = NULL;


}

TItem::~TItem () {

/* ------------------------------------- */
/* The destructor for a dictionary entry */
/* ------------------------------------- */

	debug (form ("Deallocating: %s", SName));
	// Remove from the child list
	if (lnext != NULL) lnext->lprev = lprev;
	if (lprev != NULL) lprev->lnext = lnext;

	// And from the hash list
	if (hnext != NULL) hnext->hprev = hprev;
	hprev->hnext = hnext;

	// Deallocate strings
	delete (SName);
	if (BName != NULL) delete (BName);
	if (NName != NULL) delete (NName);
}

static  TItem   *tmpit;         // To reduce the size of recursion stack frame

static  void    deallocateTree (TItem *t) {

/* ----------------------------------------------- */
/* Deallocates the ownership tree read from smurph */
/* ----------------------------------------------- */

	
	while (t != NULL) {
		if (t->ChList != NULL) deallocateTree (t->ChList);
		tmpit = t->lnext;
		delete (t);
		t = tmpit;
	}
}

#ifdef  DEBUG
void    dumpTItems ();
#endif

void    initDict () {

/* -------------------------- */
/* Initializes the dictionary */
/* -------------------------- */

	int     i;
	TItem   *ti;

	debug ("Deallocating dictionary");
#ifdef  DEBUG
	dumpTItems ();
#endif
	deallocateTree (Tree);          // Erase the ownership tree
	Tree = NULL;

	for (i = 0; i < HASHSIZE; i++) {
		// Just checking: no entry should remain if deallocateTree
		// is correct
		if (HTable [i] != NULL) {
#ifdef  DEBUG
			TItem *h;
			debug ("Garbage:");
			for (h = HTable [i]; h != NULL; h = h->hnext)
				(*dbgf) << h->SName << '\n';
#endif
		errMessage ("initDict: garbage in dictionary (internal error)");
		  HTable [i] = NULL;
		}
	}

	// Create the root entry
	Tree = ti = new TItem;
	// This one is constructed by hand
	ti->ChList = NULL;
	ti->lprev = NULL;
	ti->lnext = NULL;
	ti->BName = NULL;
	ti->NName = NULL;
	ti->SName = new char [strlen (ROOTNAME)+1];
	strcpy (ti->SName, ROOTNAME);
	// Add it to the dictionary
	HTable [i = hash (ROOTNAME)] = ti;
	ti->hnext = NULL;
	ti->hprev = (TItem*) ((&(HTable[i])) - 2);

	// OK, the rest should go automatically
}

void    initWind () {

/* --------------------- */
/* Purge the window list */
/* --------------------- */

	Window  *w, *x;

	for (w = Active; w != NULL; w = x) {
		x = w->next;
		delete (w);
	}

	Active = NULL;
}

void    initUpdt () {

/* --------------------- */
/* Purge the update list */
/* --------------------- */

	NewWindow  *w, *x;

	for (w = Updates; w != NULL; w = x) {
		x = w->next;
		delete (w);
	}

	Updates = NULL;
}

static  TItem   *objtomatch,            // To save on recursion
		*tobj;

static  TItem   *findPRec (TItem *o) {

/* ------------------------------------- */
/* A recursive searcher for 'findParent' */
/* ------------------------------------- */

	TItem   *ti;

	if (o == NULL) return (NULL);
	for (ti = o->ChList; ti != NULL; ti = ti->lnext) {
		if (ti == objtomatch) {
			return (o);
		}
		if ((tobj = findPRec (ti)) != NULL) return (tobj);
	}
	return (NULL);
}

static  char    *bntmatch;              // To save on recursion

static  TItem   *findBNam (TItem *o) {

/* -------------------------------------- */
/* A recursive searcher for 'findByBName' */
/* -------------------------------------- */

	TItem   *ti;

	if (o->BName != NULL && strcmp (o->BName, bntmatch) == 0) return (o);
	for (ti = o->ChList; ti != NULL; ti = ti->lnext)
		if (tobj = findBNam (ti)) return (tobj);
	return (NULL);
}

TItem   *findParent (TItem *o) {

/* ---------------------------------------------------------- */
/* Finds  the  parent  of an object. Note: this is a somewhat */
/* costly operation as no parent  link  is  provided  for  an */
/* object.  It  is  assumed  that this operation is performed */
/* only for backward movement  within  the  object  menu  and */
/* therefore nobody cares about its efficiency.               */
/* ---------------------------------------------------------- */

	objtomatch = o;
	return (findPRec (Tree));
}

TItem   *findByBName (char *bn) {

/* ------------------------ */
/* Finds an object by BName */
/* ------------------------ */

	bntmatch = bn;
	return (findBNam (Tree));
}

char    *TItem::dName () {

/* ------------------------------------------ */
/* Returns the displayable name of the object */
/* ------------------------------------------ */

	int     n, k, m;

	// The first part: determine when the relevant part of standard name
	// ends

	for (n = 0; SName [n] != ' ' && SName [n] != '\0'; n++);
	if (SName [n] != '\0') {
		// Locate the number part
		for (k = n+1; SName [k] == ' ' && SName [k] != '\0'; k++);
		if (SName [k] != '\0') {
			// Skip the number
			for (n = k+1; SName [k] != ' ' && SName [n] != '\0';
				n++);
			// That's it, stop right here. This way we ignore the
			// station part of a port name
		}
	}

	if (NName == NULL) {
		if (n > 80) {
			strncpy (CBuffer, SName, 77);
			CBuffer [77] = CBuffer [78] = CBuffer [79] = '.';
			CBuffer [80] = '\0';
			return (CBuffer);
		}
		if (SName [n] == '\0')          // The most often case
			return (SName);
		else {
			strncpy (CBuffer, SName, n);
			CBuffer [n] = '\0';
			return (CBuffer);
		}
	}

	// Nickname present, put it first and follow by standard name

	for (k = 0; k < 80 && NName [k] != '\0'; k++) CBuffer [k] = NName [k];
	if (k == 80) {
		CBuffer [77] = CBuffer [78] = CBuffer [79] = '.';
		CBuffer [80] = '\0';
		return (CBuffer);
	}
	if (k > 70) {
		// No room for standard name
		CBuffer [k] = '\0';
		return (CBuffer);
	}
	CBuffer [k++] = ' ';
	CBuffer [k++] = '(';
	for (m = 0; m < n && k < 79; k++, m++) CBuffer [k] = SName [m];
	if (m < n) {
		// Incomplete
		CBuffer [76] = CBuffer [77] = CBuffer [78] = '.';
		k = 79;
	}
	CBuffer [k++] = ')';
	CBuffer [k] = '\0';
	return (CBuffer);
}

char    *TItem::tName (char *sn) {

/* ------------------------------------------------- */
/* Returns the object's type name (base or standard) */
/* ------------------------------------------------- */

	int     n;

	if (sn == NULL) return (NULL);
	for (n = 0; sn [n] != ' ' && sn [n] != '\0'; n++);
	if (sn [n] == '\0') return (sn);
	if (n >= SANE) n = SANE-1;
	strncpy (CBuffer, sn, n);
	CBuffer [n] = '\0';
	return (CBuffer);
}

/* ---------------------------------------------------------- */
/* AT&T  Cfront  doesn't like loops in inlines. The following */
/* two functions are moved from types.h                       */
/* ---------------------------------------------------------- */

#if     ZZ_ATT
int  hash (register char *t) {

/* ------------------------------------ */
/* The hash function for standard names */
/* ------------------------------------ */

	register int val;

	for (val = 0; *t != '\0'; t++)
		val = ((val << 3) + *t) & (HASHSIZE-1);
	return (val);
};

TItem   *findObj (char *t) {

/* --------------------------------------- */
/* Find the object given its standard name */
/* --------------------------------------- */

	register TItem  *p;

	for (p = HTable [hash (t)]; p != NULL; p = p->hnext)
		if (strcmp (p->SName, t) == 0)
			return (p);
	return (NULL);
};
#endif
