/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ---------------------------------------------------------- */
/* This  file  contains  functions for accessing the terminal */
/* screen                                                     */
/* ---------------------------------------------------------- */

#ifdef  sgi
#define SYSV
#endif

#include        <sys/param.h>
#include        <fcntl.h>
#include        <sys/types.h>
#include        <sys/file.h>
#include        <sys/socket.h>
#include        <ctype.h>
#include        <signal.h>
#undef          CDELAY                  // Arrggghhhh !
#undef          ctob
#include        "types.h"


#include        "var.h"

#ifdef  SYSV
#include        <curses.h>
#include        <termio.h>
#include        <term.h>
#define TTYB    struct  termio
#define TGET    TCGETA
#define TSET    TCSETAF
extern "C" {
int     tgetent (char*, const char*);
int     tgetnum (const char*);
}
#else
#include        <sys/time.h>
#include        <sgtty.h>
#define TTYB    struct  sgttyb
#define TGET    TIOCGETP
#define TSET    TIOCSETP
#endif

#ifdef	ZZ_ATT_SUN
typedef	int (*OUTC) (char);
extern "C" {
char	*tgetstr (const char*, char**);
int	tgetent (char*, const char*);
int	tgetnum (const char*);
int	tputs (const char*, int, OUTC);
char	*tgoto (const char*, int, int);
}
#endif

static  TTYB    STTYStat;               // Saved terminal status

static  void    sigio (), sigttin ();

static  char    xd_is [80], xd_rs [30], xd_cl [30], xd_cm [30], xd_mr [30],
			xd_mb [30], xd_md [30], xd_me [30];

static  int     CBreak = 0,             // Flag == in cbreak mode
		dtsize;                 // Descriptor table size

void    clearScreen ();
static  void    enterCBreakMode ();
static  void    eField ();

static  char    *LUC, *RUC, *LLC, *RLC; // Corners
static  int     LUL, RUL, LLL, RLL;     // Lengths of the stuff in corners

static  int     cwx, cwy, cwxs, cwys;   // Current window parameters

static  void    sField (int);
static  void    bOut (int, int, int);

void    setCurrentWindow () {           // Overloaded a few times

	cwx = cwy = 0;                  // Full screen
	cwxs = NCols;
	cwys = NRows;
}

void    setCurrentWindow (int x, int y, int xs, int ys) {

	cwx = x;  cwy = y;  cwxs = xs + x;  cwys = ys + y;
}

void    setCurrentWindow (Window *w) {

	cwx = w->x;
	cwy = w->y;
	cwxs = w->xs + cwx;
	cwys = w->ys + cwy;
}

void    excptn (char *t) {

/* ------------- */
/* Process error */
/* ------------- */

	closeDisplay ();
	cerr << "DSD aborted: " << t << '\n';
	exit (1);
}

static  mgetstr (char *s, char **w)

/* ---------------------------------- */
/* A postprocessed version of tgetstr */
/* ---------------------------------- */

{
char    *so, *de;

	so = de = *w;
	if (tgetstr (s, w) == NULL) {   /* No such capability */
		*de = '\0';
		return (NO);
	}
	return (YES);
}

void    initDisplay () {

/* ================== */
/* Initialize display */
/* ================== */

	char    *tn,                    // Terminal name pointer
		bp [2048],              // For termcap string
		*tcp;

	int     i;

	TDesc = 0;                      // Assume that standard input is the
					// terminal
	if (!isatty (TDesc))
		excptn ("standard input must be associated with a terminal");      

	if ((tn = getenv ("TERM")) == NULL)
		excptn ("cannot determine terminal type");

	if ((i = tgetent (bp, tn)) < 0)
		excptn ("cannot open termcap file");

	if (i == 0)
		excptn ("terminal description not found in termcap database");

	// Assume some nonsensical defaults, if the size is not known
	if ((NCols = tgetnum ("co")) < 0) NCols = 80;
	if ((NRows = tgetnum ("li")) < 0) NRows = 16;

	tcp = xd_cm;
	if (mgetstr ("cm", &tcp) == NO)
		excptn ("the terminal has no cursor addressing capability");

	tcp = xd_is;
	mgetstr ("is", &tcp);
	tcp = xd_rs;
	mgetstr ("rs", &tcp);
	tcp = xd_cl;
	mgetstr ("cl", &tcp);
	tcp = xd_mr;
	mgetstr ("mr", &tcp);
	tcp = xd_mb;
	mgetstr ("mb", &tcp);
	if (strcmp (tn, "xterm") == 0) {
		// Don't use bright mode if running from X-Windows. Some idiot
		// made the size of bright characters different from the rest.
		xd_md [0] = '\0';
	} else {
		tcp = xd_md;
		mgetstr ("md", &tcp);
	}
	tcp = xd_me;
	mgetstr ("me", &tcp);

	// Reset the screen

	clearScreen ();
	enterCBreakMode ();

	// Get descriptor table size for select
	dtsize = getdtablesize ();

}

static  void    enterCBreakMode () {

/* ---------------------------------------------------------- */
/* Enters the cbreak mode. All terminal input is done in this */
/* mode without explicit echoeing.                            */
/* ---------------------------------------------------------- */

	TTYB    NewTTYStat;

	if (ioctl (TDesc, TGET, (char*)(&STTYStat)) < 0)
		excptn ("cannot get terminal status");

	NewTTYStat = STTYStat;
#ifdef  SYSV
	NewTTYStat.c_lflag &= ~ECHO & ~ECHOE & ~ECHOK & ~ECHONL & ~ICANON;
	NewTTYStat.c_oflag &= ~OPOST;
	NewTTYStat.c_cc [VEOF] = 1;     // Immediately read any single char
	NewTTYStat.c_cc [VEOL] = 0;     // No delay
#else
	NewTTYStat.sg_flags &= ~ECHO;
	NewTTYStat.sg_flags |= CBREAK;
#endif
	if (ioctl (TDesc, TSET, (char*)(&NewTTYStat)) < 0)
		excptn ("cannot enter cbreak mode");

	CBreak = YES;
}

static  void    exitCBreakMode () {

/* ------------------------------------------ */
/* Resets the terminal to the previous status */
/* ------------------------------------------ */

	if (!CBreak) return;

	ioctl (TDesc, TSET, (char*)(&STTYStat));
	CBreak = NO;
}

void    closeDisplay () {

/* ---------------------------------------- */
/* Called just before the server terminates */
/* ---------------------------------------- */

	if (CBreak) clearScreen ();
	exitCBreakMode ();
}

static  outTerm (char c) {

/* ------------------------ */
/* Terminal output function */
/* ------------------------ */

#if     ZZ_ATT
	return
#endif
	write (TDesc, &c, 1);
}

void    absMove (int col, int row) {

/* ------------------------ */
/* Absolute cursor movement */
/* ------------------------ */

	tputs (tgoto (xd_cm, col, row), 1, outTerm);
	CurCol = col;
	CurRow = row;
}

void    moveCursor (int col, int row) {

/* ------------------------------- */
/* Window-relative cursor movement */
/* ------------------------------- */

	if (col+cwx >= cwxs || row+cwy >= cwys) {
		// Pretend that the cursor is there, to avoid writing
		// outside the window
		CurCol = cwxs;
		CurRow = cwys;
	} else {
		// OK, still within window
		absMove (col + cwx, row + cwy);
	}
}

void clearScreen () {

/* -------------------------- */
/* Clears the terminal screen */
/* -------------------------- */

	tputs (xd_cl, NRows, outTerm);
	tputs (xd_rs, NRows, outTerm);
	absMove (0, 0);
}

static  inline  void    outC (char c) {

/* ------------------------------------------------ */
/* Output a single character at the curent location */
/* ------------------------------------------------ */

	write (TDesc, &c, 1);
}

void    outChar (char c) {

/* ---------------------------------------------------------- */
/* Output  a  single  character  at  the curent location with */
/* boundary checking                                          */
/* ---------------------------------------------------------- */

	if (CurRow < cwys && CurCol < cwxs) {
		outC (c);
		CurCol++;
	}
}

int     inChar () {

/* ------------------------------------------ */
/* Input a single character from the terminal */
/* ------------------------------------------ */

	char    c;

	read (TDesc, &c, 1);
	return (((int) c) & 0377);
}

char    *inNumber (char *prompt, int col, int &ex) {

/* ------------------------------------------- */
/* Read a nonnegative number from the LL field */
/* ------------------------------------------- */

	int     ic, ep, pc, bi, ft;
	char    *cp;

	absMove (0, NRows-1);
	sField (DM_REVERSE);
	for (cp = prompt, ep = 0; *cp != '\0'; cp++, ep++) outC (*cp);
	absMove (col, NRows-1);
	
	for (ft = 1, pc = col, bi = 0; ; ) {

		ic = inChar ();         // Next user character
		if (ic == '\n') {
			// Exit
			bOut (0, NRows-1, ep);  // Cleanup
			if (bi == 0) {
				// Default
				ex = NO;
				return (NULL);
			}
			CBuffer [bi] = '\0';
			ex = YES;
			return (CBuffer);
		} else if (isErase (ic)) {
			// Backspace one character
			if (bi == 0) continue;
			bi--;
			pc--;
			bOut (pc, NRows-1, ep - pc);
			ep = pc;
			absMove (pc, NRows-1);
		} else if (isKill (ic)) {
			// Abort
			bOut (0, NRows-1, ep);
			ex = NONE;
			return (NULL);
		} else if (isdigit (ic) && pc < NCols-1) {
			// Put the digit into buffer
			CBuffer [bi++] = ic;
			absMove (pc, NRows-1);
			outC (ic);
			pc++;
			if (ft) {
				// Erase the default
				ft = 0;
				bOut (pc, NRows-1, ep);
				absMove (pc, NRows-1);
			}
			if (pc > ep) ep = pc;
		}
	}
}

int     isKill (char c) {

/* ----------------------------------------- */
/* Returns 1 if c is the line kill character */
/* ----------------------------------------- */

#ifdef  SYSV
	return (STTYStat.c_cc [VKILL] == c);
#else
	return (STTYStat.sg_kill == c);
#endif
}

int     isErase (char c) {

/* ------------------------------------- */
/* Returns 1 if c is the erase character */
/* ------------------------------------- */

#ifdef  SYSV
	return (STTYStat.c_cc [VERASE] == c);
#else
	return (STTYStat.sg_erase == c);
#endif
}

inline  static  void    sField (int ftype) {

/* ---------------------------------------------------------- */
/* Output  the  sequence  to  start  a  field with a specific */
/* display attribute                                          */
/* ---------------------------------------------------------- */

	switch (ftype) {

		case    DM_REVERSE:

			if (xd_mr [0] == '\0') {
				eField ();
				return;
			}
			tputs (xd_mr, 1, outTerm);
			return;

		case    DM_BLINKING:

			if (xd_mb [0] == '\0') {
				eField ();
				return;
			}
			tputs (xd_mb, 1, outTerm);
			return;

		case    DM_BRIGHT:

			if (xd_md [0] == '\0') {
				eField ();
				return;
			}
			tputs (xd_md, 1, outTerm);
			return;

		default:

			eField ();
	}
}

void    startField (int ftype) {

/* ------------------------------------ */
/* Start a field with boundary checking */
/* ------------------------------------ */

	if (CurRow < cwys && CurCol < cwxs) sField (ftype);
}

static  inline  void    eField () {

/* ----------------- */
/* Terminate a field */
/* ----------------- */

	tputs (xd_me, 1, outTerm);
}

void    endField () {

/* ---------------------------------------- */
/* Terminate a field with boundary checking */
/* ---------------------------------------- */

	if (CurRow < cwys && CurCol <= cwxs) eField ();
}

void    outText (char *t) {

/* ----------------------------------- */
/* Output text at the current location */
/* ----------------------------------- */

	endField ();
	if (CurRow >= cwys) return;             // Outside of window
	while (CurCol < cwxs && *t != '\0') {
		write (TDesc, t++, 1);
		CurCol++;
	}
}

void    outText (int mode, char *t) {

/* ----------------------------------------------------------------- */
/* Output text at the current location in the specified display mode */
/* ----------------------------------------------------------------- */

	startField (mode);
	if (CurRow >= cwys) return;             // Outside of window
	while (CurCol < cwxs && *t != '\0') {
		write (TDesc, t++, 1);
		CurCol++;
	}
	endField ();
}

void    outText (int mode, char *t, int l) {

/* ---------------------------------------------------------- */
/* Output  text  at  the  current  location  in the specified */
/* display mode with exactly the specified length             */
/* ---------------------------------------------------------- */

	startField (mode);
	if (CurRow >= cwys) return;             // Outside of window
	while (CurCol < cwxs && l-- > 0) {
		write (TDesc, t++, 1);
		CurCol++;
	}
	endField ();
}

static  void    bOut (int c, int r, int l) {

/* ------------------------ */
/* Blank out a part of line */
/* ------------------------ */

	absMove (c, r);
	while (l--) outC (' ');
}

void    blankOut (int c, int r, int l) {

/* ------------------------------------------ */
/* Blank out a part of line (window relative) */
/* ------------------------------------------ */

	moveCursor (c, r);
	while (l--) outChar (' ');
}

/* --------------------- */
/* Operations on corners */
/* --------------------- */

static  void    dspCorner (int c, int r, int l, char *t) {

/* ---------------------------------- */
/* Display a piece of text in reverse */
/* ---------------------------------- */

	absMove (c, r);
	sField (DM_REVERSE);
	while (l--) outC (*t++);
	eField ();
}

void    clearLUC (int er) {       // Clear left upper corner

	if (LUC != NULL) {
		bOut (0, 0, LUL);
		if (er) {
			delete (LUC);
			LUC = NULL;
		}
	}
	LUL = 0;
}

void    clearRUC (int er) {       // Clear right upper corner

	if (RUC != NULL) {
		bOut (NCols-RUL, 0, RUL);
		if (er) {
			delete (RUC);
			RUC = NULL;
		}
		RUL = 0;
		if (LUC != NULL) outLUC (LUC);
	}
}

void    clearLLC (int er) {       // Clear left lower corner

	if (LLC != NULL) {
		bOut (0, NRows-1, LLL);
		if (er) {
			delete (LLC);
			LLC = NULL;
		}
	}
	LLL = 0;
}

void    clearRLC (int er) {       // Clear right lower corner

	if (RLC != NULL) {
		// Never touch the rightmost lowest character position
		bOut (NCols-RLL-1, NRows-1, RLL);
		if (er) {
			delete (RLC);
			RLC = NULL;
		}
		RLL = 0;
		if (LLC != NULL)
			outLLC (LLC);
	}
}

void    outLUC (char *t) {              // Display in left upper corner

	int     l, room;

	l = strlen (t);
	if (t != LUC) {
		clearLUC ();
		LUC = new char [l + 1];
		strcpy (LUC, t);
	} else
		clearLUC (NO);
	// Right corners have priority
	if ((LUL = l) > (room = NCols - RUL - 1)) LUL = room;
	dspCorner (0, 0, LUL, t);
}

void    outRUC (char *t) {              // Display in right upper corner

	int     l, room;

	l = strlen (t);
	if (t != RUC) {
		clearRUC ();
		RUC = new char [l + 1];
		strcpy (RUC, t);
	} else
		clearRUC (NO);
	if ((RUL = l) > NCols) RUL = NCols;
	if (RUL > (room = NCols - LUL - 1)) {
		// Partially erase the LUC stuff
		if (RUL < NCols) {
			bOut (NCols - RUL - 1, 0, 1);
			LUL = NCols - RUL - 1;
		} else
			LUL = 0;
	}
	dspCorner (NCols - RUL, 0, RUL, t);
}

void    outLLC (char *t) {              // Display in left lower corner

	int     l, room;

	l = strlen (t);
	if (t != LLC) {
		clearLLC ();
		LLC = new char [l + 1];
		strcpy (LLC, t);
	} else
		clearLLC (NO);
	// Right corners have priority
	if ((LLL = l) > (room = NCols - RLL - 1)) LLL = room;
	dspCorner (0, NRows-1, LLL, t);
}

void    outRLC (char *t) {              // Display in right lower corner

	int     l, room;

	l = strlen (t);
	if (t != RLC) {
		clearRLC ();
		RLC = new char [l + 1];
		strcpy (RLC, t);
	} else
		clearRLC (NO);
	if ((RLL = l) > NCols-1) RLL = NCols-1;
	if (RLL > (room = NCols - LLL - 2)) {
		// Partially erase the LLC stuff
		if (RLL < NCols-1) {
			bOut (NCols - RLL - 2, 0, 1);
			LLL = NCols - 1 - 2;
		} else
			LLL = 0;
	}
	dspCorner (NCols - RLL - 1, NRows-1, RLL, t);
}

void    refreshCorners () {

	if (LUC != NULL) outLUC (LUC);
	if (RUC != NULL) outRUC (RUC);
	if (LLC != NULL) outLLC (LLC);
	if (RLC != NULL) outRLC (RLC);
}

int     readyParty () {

/* ---------------------------------------------------------- */
/* Waits  for  the input from user or smurph, whichever comes */
/* first                                                      */
/* ---------------------------------------------------------- */

	fd_set  dd;

	FD_ZERO (&dd);
	FD_SET (TDesc, &dd);
	FD_SET (SmrSocket, &dd);

	while (1) {

		if (select (dtsize, &dd, NULL, NULL, NULL) == 0) continue;

		if (FD_ISSET (TDesc, &dd))
			// The terminal has priority
			return (IO_USER);
		else if (FD_ISSET (SmrSocket, &dd))
			return (IO_SMURPH);
	}
}
