/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ---------------------------------- */
/* Functions for numeric input/output */
/* ---------------------------------- */

#include        "system.h"

	// The basic i/o concept is still the same as for lansf; perhaps we will
	// revise it later.

static  long            last_integer = 0;       // Last integer number read
static  double          last_real = 0.0;        // Last floating number read
static  BIG             last_big = BIG_0;       // Last BIG number read

static  int             ni_counter = 0,         // Pending count for readInteger
#if     BIG_precision != 1
			nl_counter = 0,         // ... and for readBig
#endif
			nr_counter = 0;         // ... and for readReal

/* ---------------------------------------------------------- */
/* Input   subroutines:  for  reading  numbers  and  skipping */
/* garbage                                                    */
/* ---------------------------------------------------------- */

static  char    find_number () {

/* ---------------------------- */
/* Locate the next input number */
/* ---------------------------- */

	char    k, l;
	int     ef;

	if (EndOfData) return (' ');
SKIP:
	while (ef = Inf.readable ()) {
		Inf.get (k);
		if (((k < '0') || (k > '9')) && (k != '+') && (k != '-') &&
			(k != '%')) {

			// A character to be skipped

			if (k == '*')   // Comment - skip to eol
				while (ef = Inf.readable ()) {
					Inf.get (k);
					if (k == '\n') break;
				}
			if (!ef) break;
		} else
			// An interesting character to be looked at
			break;
	}

	if (ef) {

		// No special status

		if ((k >= '0') && (k <= '9')) {
			Inf.putback (k);
		} else {
			if ((k == '+') || (k == '-')) {
				// Must be followed by a digit
				if (ef = Inf.readable ()) {
					Inf.get (l);
					if ((l < '0') || (l > '9'))
						goto SKIP;
					else
						Inf.putback (l);
				}
			}
		}
	}

	if (ef) return (k);

	if (Inf.eof ()) {
		EndOfData = YES;
		return (0);
	};

	excptn ("i/o error on input file");

	return (0);     // No way to get here
}

void    readIn (long &frs) {

/* -------------------------- */
/* Read (long) integer number */
/* -------------------------- */

	long            res;
	char            sign, k;
	int             ef;

	Assert (zz_ifpp != NULL,
		"readIn: can't read data file after the protocol has started");
	if (ni_counter) {
		ni_counter--;
		frs = last_integer;
		return;
	}

	k = find_number ();
	if (EndOfData) {
		frs = 0;
		return;
	}

	if (k == '%') {
		frs = last_integer;
		if (Inf.readable ()) {
			Inf.get (k); Inf.putback (k);
			if ((k != '+') && (k != '-')) goto I_DONE;
		} else {
			if (Inf.eof ()) {
				EndOfData = YES;
				frs = 0;
				return;
			} else
				excptn ("i/o error on input file");
		}
	} else {
		if (k < '0') Inf.putback (k);
		frs = 0;
	}

	while (YES) {
		sign = find_number ();
		if (EndOfData) goto I_DONE;
		for (res=0; ef = Inf.readable (); ) {
			Inf.get (k);
			if (k < '0' || k > '9') break;
			res = res * 10 - (k - '0');
		}
		if (!ef && !Inf.eof ()) excptn ("i/o error on input file");
		Inf.putback (k);
		frs += (sign == '-') ? res : -res;
		if ((k != '+') && (k != '-')) break;
	}
I_DONE:
	if (ef && k == '/') {
		Inf.get (k);            // Skip the slash
		for  (ni_counter = 0; ef = Inf.readable (); ) {
			Inf.get (k);
			if (k < '0' || k > '9') break;
			ni_counter = ni_counter * 10 + (k - '0');
		}
		if (ni_counter) ni_counter--;
	}

	last_integer = frs;
}

void    readIn (double &frs) {

/* ---------------------------------------------- */
/* Read the next floating point number from input */
/* ---------------------------------------------- */

	double          res, fact;
	char            sign, k;
	int             ef;

	Assert (zz_ifpp != NULL,
		"readIn: can't read data file after the protocol has started");
	if (nr_counter) {
		nr_counter--;
		frs = last_real;
		return;
	}

	k = find_number ();
	if (EndOfData) {
		frs = 0.0;
		return;
	}

	if (k == '%') {
		frs = last_real;
		if (ef = Inf.readable ()) {
			Inf.get (k); Inf.putback (k);
			if ((k != '+') && (k != '-')) goto F_DONE;
		} else {
			if (Inf.eof ()) {
				EndOfData = YES;
				frs = 0.0;
				return;
			} else
				excptn ("i/o error on input file");
		}
	} else {
		if (k < '0') Inf.putback (k);
		frs = 0.0;
	}

	while (YES) {
		sign = find_number ();  // Find beginning of the number
		if (EndOfData) goto F_DONE;

		for (res = 0.0; ef = Inf.readable (); ) {
			Inf.get (k);
			if (k < '0' || k > '9') break;
			res = res * 10.0 - (k - '0');
		}

		if (ef && k == '.') {
			fact  = 1.0;
			while (ef = Inf.readable ()) {
				Inf.get (k);
				if (k < '0' || k > '9') break;
				fact *= 0.1;
				res -= fact * (k - '0');
			}
		}

		if (ef) Inf.putback (k);
		frs += (sign == '-') ? res : -res;
		if ((k != '+') && (k != '-')) break;
	}
F_DONE:
	if (ef && k == '/') {
		Inf.get (k);            // Skip the slash
		for  (nr_counter = 0; ef = Inf.readable (); ) {
			Inf.get (k);
			if (k < '0' || k > '9') break;
			nr_counter = nr_counter * 10 + (k - '0');
		}
		if (nr_counter) nr_counter--;
	}

	if (!ef && ! Inf.eof ()) excptn ("i/o error on input file");

	last_real = frs;
}

#if     BIG_precision != 1

void    readIn (BIG &frs) {

/* ------------------------- */
/* Read (BIG) integer number */
/* ------------------------- */

	BIG             res;
	char            sign, k;
	int             ef;

	Assert (zz_ifpp != NULL,
		"readIn: can't read data file after the protocol has started");
	if (nl_counter) {
		nl_counter--;
		frs = last_big;
		return;
	}

	k = find_number ();
	if (EndOfData) {
		frs = BIG_0;
		return;
	}

	if (k == '%') {
		frs = last_big;
		if (Inf.readable ()) {
			Inf.get (k); Inf.putback (k);
			if ((k != '+') && (k != '-')) goto B_DONE;
		} else {
			if (Inf.eof ()) {
				EndOfData = YES;
				frs = BIG_0;
				return;
			} else
				excptn ("i/o error on input file");
		}
	} else {
		if (k < '0') Inf.putback (k);
		frs = BIG_0;
	}

	while (YES) {
		sign = find_number ();  // Find number on input
		if (EndOfData) goto B_DONE;

		for (res=BIG_0; ef = Inf.readable (); ) {
			Inf.get (k);
			if (k < '0' || k > '9') break;
			res = res * 10 + (k - '0');
		}

		if (ef) Inf.putback (k);

		if (sign == '-') {
			if (res > frs)
				ierror ("negative BIG number on input");
			frs -= res;
		} else {
			frs += res;
		}
		if ((k != '+') && (k != '-')) break;
	}

B_DONE:
	if (ef && k == '/') {
		Inf.get (k);            // Skip the slash
		for  (nl_counter = 0; ef = Inf.readable (); ) {
			Inf.get (k);
			if (k < '0' || k > '9') break;
			nl_counter = nl_counter * 10 + (k - '0');
		}
		if (nl_counter) nl_counter--;
	}

	if (!ef && !Inf.eof ()) excptn ("i/o error on input file");

	last_big = frs;
}

#endif

/* ------------------ */
/* Two simple kludges */
/* ------------------ */

void    readIn (int &res) { long r; readIn (r); res = (int) r; }
void    readIn (float &res) { double r; readIn (r); res = r; }

void    encodeLong (long nn, char *s, int nc) {

/* ------------------------------------------ */
/* Encode long integer with optional exponent */
/* ------------------------------------------ */

	int             i, j, sf;
	long            al;
	char            temp [40], *tm;

	// Turn the number to negative and remember the sign
	if (nn < 0) {
		sf = 1;
	} else {
		sf = 0;
		nn = -nn;
	}

	tm = &(temp [39]);

	for (j = 0; j < 40; j++) {

		if (nn == 0) {
			if (j == 0) *--tm = '0'; else break;
			continue;
		}

		al = nn;
		*--tm = '0' + (char)((((nn /= 10) * 10) - al));
	}

	if (sf) {       // Insert minus
		j++;
		*--tm = '-';
	}

	if (j <= nc) {
		while (nc-- >  j) *s++ = ' ';
		while (nc-- >= 0) *s++ = *tm++;
		*s = '\0';
		return;
	}

	if ((i = j - nc + 2) > 9)
		i = j - nc + 3;

	if (i >= j) {
		while (nc-- > 0) *s++ = '*';
		*s = '\0';
		return;
	}

	while (j-- > i) *s++ = *tm++;
	*s++ = 'E';
	if (i > 9) {
		*s++ = (j = (i/10)) + '0';
		*s++ = (i - (j * 10)) + '0';
	} else {
		*s++ = i + '0';
	}
	*s ='\0';
}

void    encodeInt (int i, char *s, int nc) {

/* ---------------------------------------------------------- */
/* Encodes  an  integer, as opposed to long. Differs from the */
/* above in the initializer.                                  */
/* ---------------------------------------------------------- */

	encodeLong ((long) i, s, nc);
}

void    print (long n, char *hdr, int nsize, int hsize) {

/* ------------------------------------------------- */
/* Prints out an integer number preceded by a header */
/* ------------------------------------------------- */

	char    f [256], *h;

	Assert (nsize >= 0 && hsize >= 0, "print: item size is negative");
	if (hdr == NULL) h = ""; else h = hdr;

	if (hsize == 0) {
		Ouf << h;
	} else {
		while (*h != '\0' && hsize--) Ouf.put (*h++);
		while (hsize-- > 0) Ouf.put (' ');
	}

	encodeLong (n, f, nsize);
	Ouf << f;
	if (hdr != NULL) Ouf << "\n";
}

void    print (double n, char *hdr, int nsize, int hsize) {

/* ----------------------------------------------- */
/* Prints out a double number preceded by a header */
/* ----------------------------------------------- */

	char    f [16], *h;

	if (hdr == NULL) h = ""; else h = hdr;

	if (hsize == 0) {
		Ouf << h;
	} else {
		while (*h != '\0' && hsize--) Ouf.put (*h++);
		while (hsize-- > 0) Ouf.put (' ');
	}

	strcpy (f, form ("%%%1d.3g", nsize));
	Ouf << form (f, n);
	if (hdr != NULL) Ouf << "\n";
}

#if     BIG_precision != 1

void    print (const BIG &n, char *hdr, int nsize, int hsize) {

/* -------------------------------------------- */
/* Prints out a BIG number preceded by a header */
/* -------------------------------------------- */

	char    f [256], *h;

	Assert (nsize >= 0 && hsize >= 0, "print: item size is negative");
	if (hdr == NULL) h = ""; else h = hdr;

	if (hsize == 0) {
		Ouf << h;
	} else {
		while (*h != '\0' && hsize--) Ouf.put (*h++);
		while (hsize-- > 0) Ouf.put (' ');
	}

	btoa (n, f, nsize);
	Ouf << f;
	if (hdr != NULL) Ouf << "\n";
}

#endif

void    print (char *n, int nsize) {

/* ------------------- */
/* Prints out a string */
/* ------------------- */

	int     l;

	Assert (nsize >= 0, "print: item size is negative");
	if (n == NULL) n = "";

	if (nsize == 0) {
		Ouf << n;
	} else {
		l = strlen (n);
		while (nsize > l) {
			Ouf.put (' ');
			nsize--;
		}
		while (*n != '\0' && nsize--) Ouf.put (*n++);
	}
}

void    zz_ptime (TIME &t, int s) {

/* ---------------------- */
/* Prints out time values */
/* ---------------------- */

	if (undef (t)) print ("undefined", s); else print (t, s);
}
