/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */



/* ---------------------- */
/* Station initialization */
/* ---------------------- */

#include        "system.h"

void    Station::zz_start () {

/* ------------------------------------------ */
/* The intermediate-level station constructor */
/* ------------------------------------------ */

	int             i;
	static  int     asize = 15;     // A small power of two - 1 (initial
					// size of zz_st)
	Station         **scratch;
	Port            *p, *q;

	Assert (!zz_flg_started,
	  "Station: cannot create new stations after the protocol has started");

	TheStation = this;

	MQHead = MQTail = NULL;         // Initialize some variables
	CWList = SWList = NULL;
	RSList = NULL;
	ChList = NULL;
	Class  = OBJ_station;
#if     ZZ_QSL
	QSLimit = MAX_long;
	QSize = 0;
#endif
	Id = sernum++;
	if (zz_st == NULL) {
		// The first time around -- create the array of stations
		zz_st = new Station* [asize];
	} else if (Id >= asize) {
		// The array must be enlarged
		scratch = new Station* [asize];
		for (i = 0; i < asize; i++)
			// Backup copy
			scratch [i] = zz_st [i];

		delete (zz_st);         // Deallocate previous array
		zz_st = new Station* [asize = (asize+1) * 2 - 1];
		while (i--) zz_st [i] = scratch [i];
		delete (scratch);
	}

	zz_st [Id] = this;
	NStations = sernum;

	Ports = NULL;

	// Move statically-defined ports to the station's list

	for (p = zz_ncport; p != NULL; p = q) {
		q = p->nextp;
		zz_ncport = p;
		p->setup ();
	}

	// Now for the buffers
	Buffers = zz_ncpframe;
	zz_ncpframe = NULL;

#if     ZZ_TOL
	// Clock tolerance parameters
	CQuality = zz_quality;
	CTolerance = zz_tolerance;
#endif

	// Add the station to the owner's list

	if (TheProcess != NULL) {
		zz_queue_head (this, TheProcess->ChList, ZZ_Object);
		zz_DADD (TheProcess, this);
		// The above if is redundant: DisplayActive can't be set
		// when a station is created
	}
}

void    Station::setQSLimit (long lim) {

/* ---------------------------------------------------------- */
/* Sets  the  limit  on  the  number  of messages that can be */
/* queued at this station awaiting  transmission.  When  this */
/* limit is reached, new generated messages will be ignored.  */
/* ---------------------------------------------------------- */

#if     ZZ_QSL
	QSLimit = lim;
#else
	lim++;
	excptn ("Station->setQSLimit illegal: smurph not created with '-q'");
#endif
}

sexposure (Station)

	sonpaper {

		sexmode (0)

			exPrint0 (Hdr);         // Processes

		sexmode (1)

			exPrint1 (Hdr);         // Packet buffers

		sexmode (2)

			exPrint2 (Hdr);         // Activities
	}

	sonscreen {

		sexmode (0)

			exDisplay0 ();          // Processes

		sexmode (1)

			exDisplay1 ();          // Buffers

		sexmode (2)

			exDisplay2 ();          // Activities
	}
	USESID;
}

void     Station::exPrint1 (char *hdr) {

/* -------------------------------------- */
/* Print information about packet buffers */
/* -------------------------------------- */

	int                             i;
	Packet                          *p;
	ZZ_PFItem                       *pfi;

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName () << ") Packet buffers:\n\n";
	}

	Ouf << "Buf           QTime           TTime Rcvr   TP    ILength" <<
		"    TLength Flags\n";

	for (i = 0, pfi = Buffers; pfi != NULL; pfi = pfi->next, i++) {

		p = pfi->buf;
		print (zz_trunc (i,3), 3);
		Ouf << ' ';
		if (flagCleared (p->Flags, PF_full)) {
			Ouf << "empty --------- ---------------- ---- ----" <<
				" ---------- ---------- -----\n";
			continue;
		}
		print (p->QTime, 15); Ouf << ' ';
		print (p->TTime, 15); Ouf << ' ';
		if (isStationId (p->Receiver))
			print (zz_trunc (p->Receiver, 4), 4);
		else if (p->Receiver == NONE)
			print ("none", 4);
		else
			print ("bcst", 4);
		Ouf << ' ';
		print (zz_trunc (p->TP, 4), 4);
		Ouf << ' ';
		print (zz_trunc (p->ILength, 10), 10);
		Ouf << ' ';
		print (zz_trunc (p->TLength, 10), 10);
		Ouf << ' ';
		print (p->zz_pflags (), 5);
		Ouf << '\n';
	}

	Ouf << "\n(" << getOName () << ") End of list\n\n";
}

static  Station    *tstation;      // The current station for pttrav
static  ZZ_REQUEST *trq;           // Made global to save some stack space
static  ZZ_EVENT   *tev;
static  int     tii;
static  long    tjj;

void Station::pttrav (ZZ_Object *pr) {

/* --------------------- */
/* A helper for exPrint0 */
/* --------------------- */

	for ( ; (pr != NULL) && (pr->Class != AIC_process ||
		((Process*)pr)->Owner != tstation); pr = pr->next);

	if (pr == NULL) return;

	for (tev = zz_eq; tev != zz_sentinel_event; tev = tev -> next)
		if (tev->process == pr) break;

	if (tev != zz_sentinel_event && (trq = tev->chain) != NULL) {

		for (tii = 0; ; tii++) {

			if (tii == 0)
				// First time for this process
				print (pr->getOName (), 20);
			else
				Ouf << "                    ";
			Ouf << ' ';
			print (trq->ai->getTName (), 10);
			if ((tjj = trq->ai->Id) != NONE) {
				if (trq->ai->Class == AIC_port)
					tjj = tjj & 0177777;
				Ouf << form ("/%03d ", zz_trunc (tjj, 3));
			} else
				Ouf << "     ";

			print (trq->ai->zz_eid (trq->event_id), 10);
			Ouf << ' ';
			print (((Process*)pr)->zz_sn (trq->pstate), 11);

			Ouf << ' ';
			if (undef (trq->when)) {
				print ("undefined", 15);
				Ouf << ' ';
			} else {
				print (trq->when, 15);
				if (trq->when > tev->waketime)
					Ouf << ' ';
				else if (trq == tev->chain)
					Ouf << '*';
				else
					Ouf << '?';
			}
			Ouf << '\n';

			if ((trq = trq->other) == tev->chain) break;
		}
	}

	pttrav (pr->next);
	pttrav (((Process*)pr)->ChList);
}

void     Station::exPrint0 (char *hdr) {

/* --------------------------------- */
/* Print information about processes */
/* --------------------------------- */

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName () << ") Sleeping processes:\n\n";
	}

	Ouf << "             Process         AI/Idn      Event       State"
		<< "            Time\n";

	tstation = this;
	if (zz_flg_otreeAdjusted)
		pttrav (ChList);
	else
		// Ownership not adjusted ?
		pttrav (Kernel->ChList);

	Ouf << "\n(" << getOName () << ") End of list\n\n";
}

void     Station::exPrint2 (char *hdr) {

/* --------------------------------------- */
/* Print information about link activities */
/* --------------------------------------- */

	int                             l;
	ZZ_LINK_ACTIVITY                *a, *cl;
	Port                            *pp;
	Link                            *lk;

	if (hdr != NULL) {
		Ouf << hdr;
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << '(' << getOName () << ") Activities in links";
		Ouf << ':';
	}

	Ouf << "\n\n";

	Ouf << "Port Link           STime           FTime T  Rcvr   TP"
		<< "     Length";

	for (l = 0, pp = Ports; pp != NULL; pp = pp->nextp, l++) {
		lk = pp->Lnk;
		for (cl = a = lk->Archived; ; a = a -> next) {
			if (a == NULL) {
				if (cl == lk->Alive) break;
				if ((cl = a = lk->Alive) == NULL) break;
			}
			if (a->GPort != pp) continue;

			print (zz_trunc (l, 4), 4);
			print (zz_trunc (lk->Id, 4), 5);
			Ouf << ' ';
			print (a->STime, 15);
			Ouf << ' ';
			if (def (a->FTime))
				print (a->FTime, 15);
			else
				print ("undefined", 15);
			Ouf << ' ';
			if (a->Type != JAM)
				Ouf << 'T';
			else
				Ouf << 'J';
			if (cl == lk->Archived)
				Ouf << '*';
			else
				Ouf << ' ';
			Ouf << ' ';
			if (a->Type != JAM) {
				if (isStationId (a->Pkt.Receiver))
					print (zz_trunc (a->Pkt.Receiver,
						4), 4);
				else if (a->Pkt.Receiver == NONE)
					print ("none", 4);
				else
					print ("bcst", 4);
				print (zz_trunc (a->Pkt.TP, 4), 5);
				print (zz_trunc (a->Pkt.TLength, 10), 11);
			} else
				Ouf << "---- ---- ----------";

			Ouf << '\n';
		}
	}

	Ouf << "\n(" << getOName () << ") End of list\n\n";
}

void     Station::exDisplay1 () {

/* ---------------------------------------- */
/* Display information about packet buffers */
/* ---------------------------------------- */

	int                             i, j;
	Packet                          *p;
	ZZ_PFItem                       *pfi;

	for (i = 0, pfi = Buffers; pfi != NULL; pfi = pfi->next, i++) {

		p = pfi->buf;
		display (i);            // Buffer number

		if (flagCleared (p->Flags, PF_full)) {
			display ("EMPTY");
			for (j = 0; j < 6; j++) display ("--");
			continue;
		}
		display (p->QTime);
		display (p->TTime);

		if (isStationId (p->Receiver))
			display (p->Receiver);
		else if (p->Receiver == NONE)
			display ("none");
		else
			display ("bcst");

		display (p->TP);
		display (p->ILength);
		display (p->TLength);
		display (p->zz_pflags ());
	}
}

void Station::dttrav (ZZ_Object *pr) {

/* ----------------------- */
/* A helper for exDisplay0 */
/* ----------------------- */

	for ( ; (pr != NULL) && (pr->Class != AIC_process ||
		((Process*)pr)->Owner != tstation); pr = pr->next);

	if (pr == NULL) return;

	for (tev = zz_eq; tev != zz_sentinel_event; tev = tev -> next)
		if (tev->process == pr) break;
	if (tev != zz_sentinel_event && (trq = tev->chain) != NULL) {

		for (tii = 0; ; tii++) {

			if (tii == 0)
				// First time for this process
				display (pr->getOName ());
			else
				display (' ');

			display (trq->ai->getTName ());
			if ((tjj = trq->ai->Id) != NONE) {
				if (trq->ai->Class == AIC_port)
					tjj = tjj & 0177777;
				display (tjj);
			} else
				display (' ');

			display (trq->ai->zz_eid (trq->event_id));
			display (((Process*)pr)->zz_sn (trq->pstate));
			if (undef (trq->when)) {
				display ("undefined");
				display (' ');
			} else {
				dtime (trq->when);
				if (trq->when > tev->waketime)
					display (' ');
				else if (trq == tev->chain)
					display ('*');
				else
					display ('?');
			}

			if ((trq = trq->other) == tev->chain) break;
		}
	}

	dttrav (pr->next);
	dttrav (((Process*)pr)->ChList);
}

void     Station::exDisplay0 () {

/* ---------------------------------------------------------- */
/* Display  information  about  processes  belonging  to this */
/* station                                                    */
/* ---------------------------------------------------------- */

	tstation = this;
	if (zz_flg_otreeAdjusted)
		dttrav (ChList);
	else
		// Ownership not adjusted ?
		dttrav (Kernel->ChList);
}

void     Station::exDisplay2 () {

/* ----------------------------------------- */
/* Display information about link activities */
/* ----------------------------------------- */

	int                             l;
	ZZ_LINK_ACTIVITY                *a, *cl;
	Port                            *pp;
	Link                            *lk;

	for (l = 0, pp = Ports; pp != NULL; pp = pp->nextp, l++) {
		lk = pp->Lnk;
		for (cl = a = lk->Archived; ; a = a -> next) {
			if (a == NULL) {
				if (cl == lk->Alive) break;
				if ((cl = a = lk->Alive) == NULL) break;
			}
			if (a->GPort != pp) continue;

			display (l);            // Port number
			display (lk->Id);       // Link number

			display (a->STime);

			if (def (a->FTime))
				display (a->FTime);
			else
				display ("undefined");

			if (a->Type != JAM)
				display ('T');
			else
				display ('J');

			if (cl == lk->Archived)
				display ('*');
			else
				display (' ');

			if (a->Type != JAM) {
				if (isStationId (a->Pkt.Receiver))
					display (a->Pkt.Receiver);
				else if (a->Pkt.Receiver == NONE)
					display ("none");
				else
					display ("bcst");

				display (a->Pkt.TP);
				display (a->Pkt.TLength);
			} else {
				display ("---");
				display ("---");
				display ("---");
			}
		}
	}
}

sexposure (ZZ_SYSTEM)

/* ---------------------------------------------------------- */
/* SYSTEM  station  exposure is used to print out information */
/* about network topology                                     */
/* ---------------------------------------------------------- */

	sonpaper {

		sexmode (0)

			exPrint0 (Hdr);         // Full topology

		sexmode (1)

			exPrint1 (Hdr);         // Abbreviated topology
	}

	sonscreen {

		// Absent
	}
	USESID;
}

void    ZZ_SYSTEM::exPrint0 (char *hdr) {

/* --------------------------------------------- */
/* Print information about full network topology */
/* --------------------------------------------- */

	int                             l, i, j, k;
	Station                         *s;
	Link                            *lk;
	Port                            *p;
	ZZ_PFItem                       *pfi;

	Assert (zz_flg_started,
	"SYSTEM exposure: cannot expose topology before it has been defined");

	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << "(System) Full network configuration:\n\n";
	}

	Ouf << "Stations:\n\n";
	Ouf << "  NumId             Type         NPorts       NBuffers"
		<< "                NName\n";

	for (i = 0; i < NStations; i++) {
		print (i, 7); Ouf << ' ';
		s = idToStation (i);
		print (s->getTName (), 16);
		for (j = 0, p = s->Ports; p != NULL; p = p->nextp, j++);
		print (j, 15);
		for (j = 0, pfi = s->Buffers; pfi != NULL; pfi = pfi->next,j++);
		print (j, 15); Ouf << ' ';
		if (s->getNName () == NULL)
			print ("-----", 20);
		else
			print (s->getNName (), 20);
		Ouf << '\n';
	}

	Ouf << "\nLinks:\n\n";

	for (i = 0; i < NLinks; i++) {

		lk = idToLink (i);
		Ouf << form ("  Link %4d,    Type: %s\n\n", i, lk->getTName ());
		Ouf << "    Distance matrix:\n\n";

		for (j = 0; j < lk->NPorts; j++) {      // Go thru the rows
			for (k = 0; k < NStations; k++) {
				s = idToStation (k);
				for (p = s->Ports; p != NULL; p = p->nextp)
					if (p->Lnk == lk &&
						p->LRId == j) goto PFND;
			}
		excptn ("System->exposure: internal error -- port not found");
PFND:
			Ouf << form (" %5d: ", zz_trunc (j, 5));
			for (k = 0; k < lk->NPorts; k++) {
				if (k && k % 8 == 0)
					Ouf << "\n   ...: ";
				if (k == j)
					Ouf << "     ---";
				else if (k < j && lk->Type >= LT_unidirectional)
					Ouf << "     ***";
				else
					print (p->DV [k], 8);

				if ((k + 1) % 8 != 0 && (k + 1) < lk->NPorts)
					Ouf << " ";
			};
			Ouf << '\n';
			if ((j+1) < lk->NPorts && lk->NPorts > 8) Ouf << '\n';
		}
		Ouf << "\n    Ports:\n\n";
		Ouf << "      LRId    Station    SRId        Rate\n";

		for (j = 0; j < lk->NPorts; j++) {
			for (k = 0; k < NStations; k++) {
				s = idToStation (k);
				for (l = 0, p = s->Ports; p != NULL;
				    p = p->nextp, l++)
					if (p->Lnk == lk &&
						p->LRId == j) goto PFNE;
			}
		excptn ("System->exposure: internal error -- port not found");
PFNE:
			Ouf << form ("     %5d %10d %7d ", zz_trunc (j, 5),
				k, l);
			print (p->TRate, 11);
			Ouf << '\n';
		}
		Ouf << '\n';
	}
					
	Ouf << "(System) End of list\n\n";
}

void    ZZ_SYSTEM::exPrint1 (char *hdr) {

/* ---------------------------------------------------- */
/* Print abbreviated information about network topology */
/* ---------------------------------------------------- */

	int                             l, i, j, k;
	Station                         *s;
	Link                            *lk;
	Port                            *p;
	DISTANCE                        d;
	ZZ_PFItem                       *pfi;

	Assert (zz_flg_started,
	"SYSTEM exposure: cannot expose topology before it has been defined");
	if (hdr != NULL) {
		Ouf << hdr << "\n\n";
	} else {
		Ouf << "Time: "; print (Time, 15); Ouf << "     ";
		Ouf << "(System) Abbreviated network configuration:\n\n";
	}

	Ouf << "Stations:\n\n";
	Ouf << "  Numid             Type         NPorts       NBuffers"
		<< "                NName\n";

	for (i = 0; i < NStations; i++) {
		print (i, 7); Ouf << ' ';
		s = idToStation (i);
		print (s->getTName (), 16);
		for (j = 0, p = s->Ports; p != NULL; p = p->nextp, j++);
		print (j, 15);
		for (j = 0, pfi = s->Buffers; pfi != NULL; pfi = pfi->next,j++);
		print (j, 15); Ouf << ' ';
		if (s->getNName () == NULL)
			print ("-----", 20);
		else
			print (s->getNName (), 20);
		Ouf << '\n';
	}

	Ouf << "\nLinks:\n\n";

	for (i = 0; i < NLinks; i++) {

		lk = idToLink (i);

		// Determine the length of the link
		for (d = DISTANCE_0, k = 0; k < NStations; k++) {
			s = idToStation (k);
			for (p = s->Ports; p != NULL; p = p->nextp)
				if (p->Lnk == lk &&
					p->MaxDistance > d) d = p->MaxDistance;
		}

		Ouf << form ("  Link %4d,    Type: %s,    Length: ", i,
			lk->getTName ());
		print (d, 11);
		Ouf << "\n\n";

		Ouf << "    Ports:\n\n";
		Ouf << "      LRId    Station    SRId        Rate\n";

		for (j = 0; j < lk->NPorts; j++) {
			for (k = 0; k < NStations; k++) {
				s = idToStation (k);
				for (p = s->Ports, l = 0; p != NULL;
				    p = p->nextp, l++)
					if (p->Lnk == lk &&
						p->LRId == j) goto PFNF;
			}
		excptn ("System->exposure: internal error -- port not found");
PFNF:
			Ouf << form ("     %5d %10d %7d ", zz_trunc (j, 5),
				k, l);
			print (p->TRate, 11);
			Ouf << '\n';
		}
		Ouf << '\n';
	}

	Ouf << "(System) End of list\n\n";
}
