/* oooooooooooooooooooooooooooooooooo */
/* Copyright (C) 1991   P. Gburzynski */
/* oooooooooooooooooooooooooooooooooo */

/* --- */

/* ------------------------------------------------------ */
/* This is a simple interactive initializer of everything */
/* ------------------------------------------------------ */

#include        "version.h"

#include        <stream.h>
#ifdef	ZZ_ATT
#include	<string.h>
#include	<stdlib.h>
#include	<osfcn.h>
#include	<sys/fcntl.h>
extern "C" {
char *getwd (char*);
}
#define	FLUSH	<< flush
#else
#define	FLUSH
#endif
#include        <sys/param.h>
#include        <sys/types.h>
#include        <sys/socket.h>
#include        <netinet/in.h>
extern "C" {
#include        <netdb.h>
}
#include        <ctype.h>

#include        "defaults.h"

#define	BFFSIZE	8192

char    is [BFFSIZE],
	cb [BFFSIZE];

char    wd [MAXPATHLEN];

char    *incdir, *ccomp, *srcdir, *libdir, *monhost, *mkname, *mksdir;
int     maxlib, monport, ownlib;

#ifndef	ZZ_ATT
int	FError;		// File error flag

void	fError (char *t) {

// File error handler

	FError = 1;
}
#endif

ostream	*openOStream (char *fn) {

// Opens an output stream for a given file name

#if	ZZ_ATT

	int	fd;

	fd = open (fn, O_WRONLY | O_CREAT | O_TRUNC, 0666);
	if (fd < 0) return (NULL);
	return (new ostream (fd));
#else
	ostream *os;
	set_File_error_handler (fError);
	FError = 0;
	os = new ostream (fn, "w");
	return (FError ? NULL : os);
#endif
}

istream	*openIStream (char *fn) {

// Opens an input stream for a given file name

#if	ZZ_ATT

	int	fd;

	fd = open (fn, O_RDONLY, 0);
	if (fd < 0) return (NULL);
	return (new istream (fd));
#else
	istream *is;
	set_File_error_handler (fError);
	FError = 0;
	is = new istream (fn, "r");
	return (FError ? NULL : is);
#endif
}

void    getstring () {

/* ----------------------------------------------- */
/* Reads a string of characters from 'cin' to 'is' */
/* ----------------------------------------------- */

	int     i;
	char    c;

	if (cin.eof ()) {

		// Keep returning empty string on eof
		is [0] = '\0';
		return;
	}

	while (1) {

		cin.get (c);
		if (!isspace (c)) break;
		if (c == '\n') {
			is [0] = '\0';
			return;
		}
	}

	for (is [0] = c, i = 1; i < 8192; i++) {

		cin.get (c);
		if (isspace (c)) break;
		is [i] = c;
	}

	is [i] = '\0';

	while (c != '\n' && ! cin.eof ()) cin.get (c);
}

void	encint (int v) {

/* ---------------------------------------------- */
/* Encodes a nonnegative integer number into 'is' */
/* ---------------------------------------------- */

	char  code [16];
	int   n;

	for (n = 14; ; n--) {
		code [n] = (v % 10) + '0';
		if ((v = v / 10) == 0) break;
	}
	code [15] = '\0';
	strcpy (is, & (code [n]));
}
	
main () {

	char    c, *s;
	istream *vfi;
	ostream	*vfo;
	int	i, m;

cout << "Hi, this is SMURPH Version " << VERSION << '\n';
cout << "You will be asked a few simple questions.  By  hitting  RETURN you\n";
cout << "select the default answer indicated in the question.  Note that in\n";
cout << "the vast majority of cases the default answers are fine.\n\n";

#if	ZZ_ATT
cout << '\n';
cout << "Note: With the AT&T cfront compiler,  SMURPH will be using its own\n";
cout << "version    of    the   stream   i/o   library   and   the   memory\n";
cout << "allocator/deallocator.    The SMURPH version of the stream library\n";
cout << "contains only a few basic tools. Please use the i/o tools provided\n";
cout << "by the package.\n\n";
#endif

	cout << "What is the name of your C++ compiler? (" << DEFCCOMP << "): ";

	getstring ();

	if (is [0] == '\0') {
		cout << "Assuming " << DEFCCOMP << '\n';
		strcpy (is, DEFCCOMP);
	} else
		cout << "The name of your compiler is " << is << '\n';

	ccomp = new char [strlen (is) + 1];
	strcpy (ccomp, is);

cout << '\n';
cout << "Give  me  the  path  to  SMURPH  sources  (at least from your home\n";
cout << "directory).  If the path starts with '/', it will be assumed to be\n";
cout << "absolute;  otherwise, it will be interpreted relative to your home\n";
cout << "directory. The default path is:\n\n";

	getwd (wd);
	cout << "        " << wd << "\n\n";

cout << "which is the current working directory.\n";

SP_RETRY:

	getstring ();

	if (is [0] == '\0') {
		cout << "Assuming " << wd << '\n';
		strcpy (is, wd);
	} else {
		// Check if the directory is reachable
		if (is [0] != '/') {
			// Turn it into $HOME/is
			if ((s = getenv ("HOME")) == NULL) {
				cout << "Cannot determine your home directory;"
					<< " please specify the full path\n";
				goto SP_RETRY;
			}
			strcpy (is+4096, is);
			strcpy (is, s);
			strcat (is, "/");
			strcat (is, is+4096);
		}

		// Execute a tentative chdir there

		if (chdir (is) < 0) {
			cout << "This directory is inaccessible;" <<
				" please specify another path\n";
			goto SP_RETRY;
		}
		// Move back to where you should be
		chdir (wd);
	}

	srcdir = new char [strlen (is) + 1];
	strcpy (srcdir, is);

#if ZZ_ATT
	ownlib = 1;
#else
cout << '\n';
cout << "Do  you  want  SMURPH  to  use  its own version of i/o library and\n";
cout << "memory allocator/deallocator? If you answer 'no', the standard i/o\n";
cout << "library and memory allocator  (malloc) will be used.   The  SMURPH\n";
cout << "version  (default)  is  more  efficient,  but it only contains the\n";
cout << "essential subset of the i/o operations. (y)\n";
	getstring ();
	ownlib = (is [0] == 'y' || is [0] == 'Y' || is [0] == '\0');
	cout << "Assuming " << (ownlib ? "private" : "standard") <<
		" library\n";
#endif
cout << '\n';
cout << "Please give the path to the  directory  where  you  want  to  keep\n";
cout << "binary libraries.  This directory need not exist at present; if it\n";
cout << "exists, however, IT WILL BE CLEARED. The default path is:\n\n";

	for (s = wd + strlen (wd); s != wd && *(s-1) != '/'; s--);
	*s = '\0';
	strcat (wd, DEFLIBDR);
	cout << "        " << wd << "\n\n";

cout << "which is equivalent to: ../" << DEFLIBDR << '\n';

LB_RETRY:

	getstring ();

	if (is [0] == '\0') {
		cout << "Assuming " << wd << '\n';
		strcpy (is, wd);
	} else {
		if (is [0] != '/') {
			// Turn it into $HOME/is
			if ((s = getenv ("HOME")) == NULL) {
				cout << "Cannot determine your home directory;"
					<< " please specify the full path\n";
				goto LB_RETRY;
			}
			strcpy (is+4096, is);
			strcpy (is, s);
			strcat (is, "/");
			strcat (is, is+4096);
		}
	}

	// Check if the directory exists
	getwd (wd);             // Save current wd
	if (chdir (is) < 0) {

		// Assume that it does not exist and try to create it

		for (s = is+1; *s != '\0'; s++) {
			if (*s == '/' && *(s+1) != '\0') {
				c = *s;
				*s = '\0';
				mkdir (is, 0777);
				*s = c;
			}
		}

		if (mkdir (is, 0777) < 0) {
			cout << "I have problems accessing this";
			cout << " directory; please speify another";
			cout << " path\n";
			goto LB_RETRY;
		}

	} else {
		chdir (wd);     // Move back
	}

	// Clear the directory
	strcpy (cb, "rm -fr ");
	strcat (cb, is);
	strcat (cb, "/* 2>/dev/null");
	system (cb);
	libdir = new char [strlen (is) + 1];
	strcpy (libdir, is);

cout << '\n';
cout << "Specify the maximum number of versions of the binary library to be\n";
cout << "kept simultaneously.  Whenever the configuration of options of the\n";
cout << "SMURPH  make  program requests a new library version to be created\n";
cout << "and the current number of library version is equal to the maximum,\n";
cout << "the least recently used library will be removed. The default limit\n";
cout << "is " << DEFMAXLB << '\n';

ML_RETRY:

	getstring ();

	if (is [0] == '\0') {
		cout << "Assuming " << (maxlib = DEFMAXLB) << '\n';
	} else {
		for (s = is; *s != '\0' && isdigit (*s); s++);
		if (*s != '\0') {
			cout << "This is not a legal number; try again\n";
			goto ML_RETRY;
		}
		maxlib = atoi (is);
		cout << "The limit is " << maxlib << '\n';
	}

cout << '\n';
cout << "If  you  want  to  maintain  a  library of protocol include files,\n";
cout << "please give me the path to  that  directory;  otherwise  just  hit\n";
cout << "'return'.\n";

IN_RETRY:

	getstring ();

	if (is [0] == '\0') {
		cout << "Assuming 'no include library'\n";
		incdir = NULL;
	} else {
		if (is [0] != '/') {
			// Turn it into $HOME/is
			if ((s = getenv ("HOME")) == NULL) {
				cout << "Cannot determine your home directory;"
					<< " please specify the full path\n";
				goto IN_RETRY;
			}
			strcpy (is+4096, is);
			strcpy (is, s);
			strcat (is, "/");
			strcat (is, is+4096);
		}

		// Check if the directory exists
		getwd (wd);             // Save current wd
		if (chdir (is) < 0) {

			// Assume that it does not exist and try to create it

			for (s = is+1; *s != '\0'; s++) {
				if (*s == '/' && *(s+1) != '\0') {
					c = *s;
					*s = '\0';
					mkdir (is, 0777);
					*s = c;
				}
			}

			if (mkdir (is, 0777) < 0) {
				cout << "I have problems accessing this";
				cout << " directory; please specify another";
				cout << " path\n";
				goto IN_RETRY;
			}

		} else {
			chdir (wd);     // Move back
		}

		incdir = new char [strlen (is) + 1];
		strcpy (incdir, is);
	}

cout << '\n';
cout << "Specify  the  number  of  the monitor socket port on the host that\n";
cout << "will be running the SMURPH monitor.  The default  port  number  is\n";
cout << DEFMONPT << '\n';

PN_RETRY:

	getstring ();

	if (is [0] == '\0') {
		cout << "Assuming " << (monport = DEFMONPT) << '\n';
	} else {
		for (s = is; *s != '\0' && isdigit (*s); s++);
		if (*s != '\0') {
			cout << "This is not a legal number; try again\n";
			goto PN_RETRY;
		}
		monport = atoi (is);
	}

cout << '\n';
cout << "Specify the host running the SMURPH monitor. The default host is:\n\n";

	gethostname (wd, MAXPATHLEN-1);

	cout << "        " << wd << "\n\n";

cout << "which is the current host\n";

HS_RETRY:

	getstring ();

	if (is [0] == '\0') {
		cout << "Assuming " << wd << '\n';
		strcpy (is, wd);
	} else {
		// Check if the host is reachable
		if (gethostbyname (is) == NULL) {
			cout << "This host is unknown; please specify";
			cout << " another host\n";
			goto HS_RETRY;
		}
	}

	monhost = new char [strlen (is) + 1];
	strcpy (monhost, is);

cout << '\n';
cout << "Specify the name of the SMURPH make program (" << DEFMKNAM << "): ";

	getstring ();

	if (is [0] == '\0') {
		cout << "Assuming " << DEFMKNAM << '\n';
		strcpy (is, DEFMKNAM);
	}

	mkname = new char [strlen (is) + 1];
	strcpy (mkname, is);

cout << '\n';
cout << "Specify  the  path  to  the  directory that is to contain the make\n";
cout << "program. The default path is:\n\n";

	if ((s = getenv ("HOME")) == NULL)
		wd [0] = '\0';
	else
		strcpy (wd, s);
	strcat (wd, "/");
	strcat (wd, DEFMKDIR);

	cout << "        " << wd << "\n\n";

MD_RETRY:

	getstring ();

	if (is [0] == '\0') {
		cout << "Assuming " << wd << '\n';
		strcpy (is, wd);
	} else {
		if (is [0] != '/') {
			// Turn it into $HOME/is
			if ((s = getenv ("HOME")) == NULL) {
				cout << "Cannot determine your home directory;"
					<< " please specify the full path\n";
				goto MD_RETRY;
			}
			strcpy (is+4096, is);
			strcpy (is, s);
			strcat (is, "/");
			strcat (is, is+4096);
		}
	}

	// Execute a tentative chdir there

	getwd (wd);
	if (chdir (is) < 0) {
		cout << "This directory does not exist;";
		cout << " should I create it? (y): ";
		strcpy (wd, is);
		getstring ();
		if (is [0] != 'y' && is [0] != 'Y' && is [0] != '\0') {
			cout << "Try another path\n";
			goto MD_RETRY;
		}
		strcpy (is, wd);

		for (s = is+1; *s != '\0'; s++) {
			if (*s == '/' && *(s+1) != '\0') {
				c = *s;
				*s = '\0';
				mkdir (is, 0777);
				*s = c;
			}
		}

		if (mkdir (is, 0777) < 0) {
			cout << "I have problems accessing this";
			cout << " directory; please specify another";
			cout << " path\n";
			goto MD_RETRY;
		}
	} else {
		chdir (wd);
	}

	mksdir = new char [strlen (is) + 1];
	strcpy (mksdir, is);

	// OK, Now update the version.h file

	if ((vfi = openIStream ("version.h")) == NULL) {
COVH:
		cout << "Sorry, I can't open version.h\n";
		exit (1);
	}

	// Read everything from version.h up to the line that looks like
	// #define ZZ_SOURCES ... or to the end

#if	ZZ_ATT
	for (m = 0; m < BFFSIZE && !vfi->eof (); m++) vfi->get (cb[m]);
#else
	for (m = 0; m < BFFSIZE && vfi->readable (); m++) vfi->get (cb[m]);
#endif

	if (m >= BFFSIZE) {
		cout << "Sorry, the version.h file seems too long. Recompile\n";
		cout << "maker with increased BFFSIZE and try again.\n";
		exit (1);
	}

	// Look up #define first

	for (i = 0; i < m; i++) {
		if (strncmp (&(cb [i]), "#define", 7) == 0) {
			i += 7;
			while (i < m && isspace (cb [i])) i++;
			if (i >= m) break;
			if (strncmp (&(cb [i]), "ZZ_SOURCES", 10) == 0) {
				// Find the first newline back
				while (i >= 0 && cb [i] != '\n') i--;
				i++;
				break;
			}
		}
	}

#ifndef	ZZ_ATT
	vfi->close ();
#endif
	delete (vfi);

	// Now rewrite version.h

	vfo = openOStream ("version.h");
	if (vfo == NULL) goto COVH;

	for (m = 0; m < i; m++) vfo->put (cb [m]);

	*vfo << "#define  ZZ_SOURCES      \"" << srcdir << "\"\n";
	*vfo << "#define  ZZ_LIBPATH      \"" << libdir << "\"\n";
	*vfo << "#define  ZZ_INCPATH      \"";
	if (incdir != NULL)
		*vfo << incdir;
	else
		*vfo << '.';
	*vfo << "\"\n";
	if (ownlib)
		*vfo << "#define  ZZ_OWNLIB       1\n";
	*vfo << "#define  ZZ_MONHOST      \"" << monhost << "\"\n";
	*vfo << "#define  ZZ_MONSOCK      " << monport << '\n';

#ifndef	ZZ_ATT
	vfo->close ();
#endif
	delete (vfo);

	if (ownlib) {	// Create the private library
		getwd (wd);
		chdir ("LIB");
		cout << "Creating i/o library + malloc ...\n" FLUSH;
		strcpy (cb, "make COMP=");
		strcat (cb, ccomp);
		if (system (cb)) {
#ifdef	mips
CIOPB:
#endif
			cout << "Problems creating the library,";
			cout << " procedure aborted\n" FLUSH;
			exit (1);
		}
#ifdef	mips
		cout << "Ignore the following two assembler warnings:\n" FLUSH;
		if (system ("cc -c mips.s")) goto CIOPB;
#endif
		chdir (wd);
	}

	// Create mks

	cout << "Creating 'mks' ...\n" FLUSH;

	strcpy (cb, ccomp);
	strcat (cb, " -o ");
	strcat (cb, mksdir);
	strcat (cb, "/");
	strcat (cb, mkname);
	strcat (cb, " mks.c ");
	if (ownlib) strcat (cb, "LIB/lib.o LIB/iolink.o ");
	strcat (cb, "-DCCOMP=\\\"");
	strcat (cb, ccomp);
	strcat (cb, "\\\" -DMAXLIB=");
	encint (maxlib);
	strcat (cb, is);
	if (system (cb)) {
		cout << "Problems installing 'mks', procedure aborted\n" FLUSH;
		exit (1);
	}
	strcpy (cb, "strip ");
	strcat (cb, mksdir);
	strcat (cb, "/");
	strcat (cb, mkname);
	system (cb);

	// Create the preprocessor

	getwd (wd);
	chdir ("SMPP");
	cout << "Creating 'smpp' ...\n" FLUSH;
	strcpy (cb, "make ");
	strcat (cb, " OLPATH=");
	if (ownlib)
		strcat (cb, "\"../LIB/lib.o ../LIB/iolink.o\"");
	strcat (cb, " COMP=");
	strcat (cb, ccomp);
#ifdef	ZZ_ATT_SUN
	strcat (cb, " OPTI=-w");
#else
	strcat (cb, " OPTI=-O");
#endif
	if (system (cb)) {
		cout << "Problems creating smpp,";
		cout << " procedure aborted\n" FLUSH;
		exit (1);
	}
	system ("strip smpp");
	chdir (wd);

	cout << "Should I create the monitor ? (y): ";
	getstring ();
	if (is [0] == 'y' || is [0] == 'Y' || is [0] == '\0') {

		getwd (wd);
		chdir ("MONITOR");
		cout << "Creating 'monitor' ...\n" FLUSH;
		strcpy (cb, "make COMP=");
		strcat (cb, ccomp);
#ifdef	ZZ_ATT_SUN
		strcat (cb, " OPTI=-w");
#else
		strcat (cb, " OPTI=-O");
#endif
		if (system (cb)) {
			cout << "Problems creating monitor,";
			cout << " procedure aborted\n" FLUSH;
			exit (1);
		}
		system ("strip monitor");
		chdir (wd);

cout << '\n';
cout << "The monitor is on file 'MONITOR/monitor'. Run it in the background\n";
cout << "but first make sure that no other copy of the monitor is running.\n";
	}

cout << '\n';
	cout << "Should I create the terminal display program DSD? (y): ";
	getstring ();
	if (is [0] == 'y' || is [0] == 'Y' || is [0] == '\0') {

		getwd (wd);
		chdir ("DSD");
		cout << "Creating 'dsd' ...\n" FLUSH;
		strcpy (cb, "make ");
		strcat (cb, " OLPATH=");
		if (ownlib)
			strcat (cb, "\"../LIB/lib.o ../LIB/iolink.o\"");
		strcat (cb, " COMP=");
		strcat (cb, ccomp);
#ifdef	ZZ_ATT_SUN
		strcat (cb, " OPTI=-w");
#else
		strcat (cb, " OPTI=-O");
#endif
		if (system (cb)) {
			cout << "Problems creating dsd,";
			cout << " procedure aborted\n" FLUSH;
			exit (1);
		}

		system ("strip dsd");
		cout << '\n';
		cout << "The server is on file 'DSD/dsd'.\n";
		cout << "Should I move it to '" << mksdir << "? (y): ";
		getstring ();
		if (is [0] == 'y' || is [0] == 'Y' || is [0] == '\0') {
			cout << '\n';
			cout << "Specify the name of the server (dsd): ";
			getstring ();
			if (is [0] == '\0') strcpy (is, "dsd");
			strcpy (cb, "mv dsd ");
			strcat (cb, mksdir);
			strcat (cb, "/");
			strcat (cb, is);
			system (cb);
		}
		chdir (wd);
	}
	cout << "Done.\n" FLUSH;
}
